Unit Petri;                      { Place, Transition, Arc }
       { This unit defines generic Petri Networks objects }
{$N+,E+}

Interface

Uses Crt, Graph, GGraph, PNList, GrEditS, Randoms;

{ GGraph    = General graph }
{ PNList    = Two way linked lists (dynamic data struct.) }
{ GrEditS   = Graphical editor (objects: polyline, point) }

{***************** Public declarations *******************}

Const { Nets properties: }
      Horiz:Byte  = 0;     { Horizontal direction of a transition }
      Vert :Byte  = 1;      { Vertical  direction of a transition }
      Namelen     = 15;         { Maximum length of names }
      DPcolI      = Blue;       { Default color of places }
      DPMcolI     = Yellow;   { Default color of markings }
      DPsizeI     = 20;          { Default size of places }
      DTcolI      = Green; { Default color of transitions }
      DTsizeI     = 50;     { Default size of transitions }
      IAColI      = LightGreen;     { Color of input arcs }
      OAColI      = LightBlue;     { Color of output arcs }
      MarColI     = White;             { Arc weight color }
      TimColI     = White;        { Firing end time color }

Type  PetriPP = ^PetriPlace;        { Pointers to objects }
      PetriTP = ^PetriTransition;
      PetriAP = ^PetriArc;

     PetriPlace = Object
        Xpos,Ypos : Integer;                 { Center }
        Size : Integer;                      { Radius }
        Color : Byte;  { Color of the circle and name }
        Mcolor : Byte;         { Color of the marking }
        Marking : LongInt;         { Number of tokens }
        OldMark : LongInt;         { Recorded marking }
        Name : String[Namelen];                { Name }
        PType : Integer; { Place type - used by users }
        Timing   : Boolean;  { Whether to record tokens timing }
       {Methods:}
        Procedure PInit(Xp,Yp,Siz:Integer; Col,MCol:Byte;
                        Mar:LongInt; Nam:String; Typ:Integer);

        Procedure PSimInit;
        Procedure PShow;
        Procedure PErase;
        Procedure ShowMarking(Col:Byte);
     End{Place};

     PetriTransition = Object
        Xpos,Ypos : Integer;               { Midpoint }
        Size : Integer;                      { Length }
        Color : Word;     { Color of the bar and name }
        Direct : Byte;     { Direction: Horiz or Vert }
        Name : String[Namelen];                { Name }
        Inlist : Listpoint;  { List of Arcs from input places }
        Outlist : Listpoint;  { List of Arcs to output places }
                      { Note: Arcs contain pointers to places }
        FireEnd : Real;        { Time the firing ends }
        Firing  : Boolean;         { Firing is On/Off }
        TranNum : Integer;        { Transition Number }
        TType : Integer;    { Transition type - used by users }
        Priority : Word;   { Priority: default 100, 0 maximum }
        RandDur : Rand;      { Random firing duration }
       {Methods:}
        Procedure TInit(Xp,Yp,Siz:Integer; Col,Dir:Byte;
                        Nam:String; TNum,Typ:Integer);
        Procedure TSimInit;
        Procedure ShowShape;
        Procedure ShowStatus;
        Procedure TShow;           { Both shape & status }
        Procedure EraseShape;
        Procedure EraseStatus;
        Procedure TErase;          { Both shape & status }
        Function  Enabled:Boolean;
        Procedure StartFire;             { Remove tokens }
        Procedure EndFire;                  { Add tokens }
        Procedure Fire;           { Firing: Remove & Add }
     End{Transition};

     PetriArc = Object (Poly)      { See the unit GrEdit }
        Place : PetriPP;              { Pointer to place }
        Transition : PetriTP;    { Pointer to transition }
        Weight : LongInt;            { Edge multiplicity }
        AType : Integer;                  { Type of arc: }
           { -1=Output, 0=Normal, 1=Inhibitor, 2=Testing }
        BranchPro : Real;        { Branching probability }
        BranchCdf : Real;                { Branching Cdf }
       {Methods:}
        Procedure AInit(Col:Byte; Wid:Byte; Sty:Byte;
                        Wei:LongInt; Typ:Integer);
        Procedure ShowWeight;
        Procedure EraseWeight;
        Procedure AShow;
        Procedure AErase;
     End{Arc};
(*     Poly = object                { INHERITED }
       Corners : Listpoint;    { List of points }
       Color   : Byte;
       Width   : Byte;
       Style   : Byte;
      {Methods:}
       Constructor Init(Col:Byte;Wid:Byte;Sty:Byte);
       Procedure Create;
       Procedure Show;
       Procedure ShowWithArrow;
       Procedure Destroy;
     end;                   *)

Var   DPcol    :Byte;  { Current values: Color of places }
      DPMcol   :Byte;                { Color of markings }
      DPsize   :Integer;                { Size of places }
      DTcol    :Byte;             { Color of transitions }
      DTsize   :Integer;           { Size of transitions }
      IACol    :Byte;              { Color of input arcs }
      OACol    :Byte;             { Color of output arcs }
      MarCol   :Byte;                 { Arc weight color }
      TimCol   :Byte;            { Firing end time color }

  Procedure ShowNum(n:LongInt; X,Y:Word; Col:Byte);
                                   { Shows n as a String }

Implementation

  Procedure ShowNum(n:LongInt; X,Y:Word; Col:Byte);
                                   { Shows n as a String }
  Var s:String;
      c:Char;
      m:LongInt;
  Begin
    GSetColor(Col);
    SetTextStyle(MenuFont,HorizDir,MenuSize);
    SetTextJustify(CenterText,CenterText);
    s := '';
    m := Abs(n);
    Repeat
       c := Chr(Ord('0') + m mod 10);
       s := c + s;
       m := m div 10;
    Until m=0;
    If n<0 then s := '-' + s;
    OutTextXY(X,Y,s);
  end;

{******************** Methods for PetriPlace **********************}

  Procedure PetriPlace.PInit(Xp,Yp,Siz:Integer; Col,MCol:Byte;
                             Mar:LongInt; Nam:String; Typ:Integer);
  Begin
     Xpos := Xp;
     Ypos := Yp;
     Size := Siz;
     Color := Col;
     Mcolor := MCol;
     Marking := Mar;
     Name := Nam;
     PType := Typ;
     OldMark := Marking;
     Timing := false;
  end;

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

  Procedure PetriPlace.PSimInit;
  Begin
  End;

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

  Procedure PetriPlace.PShow;
  Begin
     GSetColor(Color);
     SetLineStyle(SolidLn,0,NormWidth);
     Circle(Xpos,Ypos,Size);
     SetTextStyle(MenuFont,HorizDir,MenuSize);
     SetTextJustify(LeftText,CenterText);
     OutTextXY(Xpos+3*Size div 4 + 2,
                Ypos+3*Size div 4 + 2, Name);
     ShowMarking(Mcolor);
  end;

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

  Procedure PetriPlace.PErase;
  Begin
     GSetColor(Black);
     SetLineStyle(SolidLn,0,NormWidth);   
     Circle(Xpos,Ypos,Size);
     SetTextStyle(MenuFont,HorizDir,MenuSize);
     SetTextJustify(LeftText,CenterText);
     OutTextXY(Xpos+3*Size div 4 + 2,
                Ypos+3*Size div 4 + 2, Name);
     ShowMarking(Black);
  end;

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

  Procedure PetriPlace.ShowMarking(Col:Byte);
  Begin
    ShowNum(Marking,Xpos,Ypos,Col);
  end;


{****************** Methods for PetriTransition *******************}

  Procedure PetriTransition.TInit(Xp,Yp,Siz:Integer; Col,Dir:Byte;
                                  Nam:String; TNum,Typ:Integer);
  Begin
     Xpos    := Xp;
     Ypos    := Yp;
     Size    := Siz;
     Color   := Col;
     Direct  := Dir;
     Name    := Nam;
     TranNum := TNum;
     TType   := Typ;
     Inlist  := Newlist;
     Outlist := Newlist;
     FireEnd := 0;             { Only value > 0 displayed ! }
     Firing  := false;                     { Default values }
     Priority := 100;
     RandDur.Init;
  end;

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

  Procedure PetriTransition.TSimInit;
  Begin
  End;

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

  Procedure PetriTransition.ShowShape;
  Begin
     GSetColor(Color);
     SetLineStyle(SolidLn,0,Thickwidth);
     SetTextStyle(MenuFont,HorizDir,MenuSize);
     If Direct=Horiz then Begin
         Line(Xpos - Size div 2,Ypos,Xpos + Size div 2,Ypos);
         SetTextJustify(LeftText,CenterText);
         OutTextXY(Xpos + Size div 2 + 6, Ypos, Name);
       end{then}
     else Begin
         Line(Xpos,Ypos - Size div 2,Xpos,Ypos + Size div 2);
         SetTextJustify(CenterText,TopText);
         OutTextXY(Xpos, Ypos + Size div 2 + 5,Name);
     end{if};
  end;

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

  Procedure PetriTransition.ShowStatus;
  Var s : String;
  Begin
     If Enabled or Firing then Begin
       If GBandW then Begin
         GSetColor(White);
         If Firing then
           SetLineStyle(DottedLn,0,NormWidth)
         else
           SetLineStyle(SolidLn,0,NormWidth);
         If Direct=Horiz then
           Rectangle(Xpos-Size div 2-3,Ypos-4,Xpos+Size div 2+3,Ypos+4)
         else
           Rectangle(Xpos-4,Ypos-Size div 2-3,Xpos+4,Ypos+Size div 2+3);
        end
       else Begin
         If Firing then
           SetFillStyle(SlashFill,Red)
         else
           SetFillStyle(SolidFill,Red);
         If Direct=Horiz then
           Bar(Xpos-Size div 2-2,Ypos-3,Xpos+Size div 2+2,Ypos+3)
         else
           Bar(Xpos-3,Ypos-Size div 2-2,Xpos+3,Ypos+Size div 2+2);
       end{if};
       If Firing then Begin
        Str(FireEnd:3:2,s);
        GSetColor(TimCol);
        SetTextStyle(MenuFont,HorizDir,MenuSize);
        SetTextJustify(CenterText,CenterText);
        If Direct=Horiz then Begin
           SetTextJustify(LeftText,BottomText);
           OutTextXY(Xpos - Size div 2 - 4, Ypos - 8, s);
          end{then}
        else Begin
           SetTextJustify(CenterText,BottomText);
           OutTextXY(Xpos, Ypos - Size div 2 - 7, s);
        end{if};
       end{if};
     end{if};
  end;

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

  Procedure PetriTransition.TShow;
  Begin
     ShowShape;
     ShowStatus;
  end;

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

  Procedure PetriTransition.ERaseShape;
  Var s : String;
  Begin
     GSetColor(Black);
     SetTextStyle(MenuFont,HorizDir,MenuSize);
     SetFillStyle(SolidFill,Black);
     If Direct=Horiz then Begin
         Bar(Xpos-Size div 2-2,Ypos-3,Xpos+Size div 2+2,Ypos+3);
         SetTextJustify(LeftText,CenterText);
         OutTextXY(Xpos + Size div 2 + 6, Ypos, Name);
       end{then}
     else Begin
         Bar(Xpos-3,Ypos-Size div 2-2,Xpos+3,Ypos+Size div 2+2);
         SetTextJustify(CenterText,TopText);
         OutTextXY(Xpos, Ypos + Size div 2 + 5, Name);
     end{if};
  end;

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

  Procedure PetriTransition.ERaseStatus;
  Var s : String;
  Begin
     If FireEnd>0 then Begin         { End of firing - if > 0 }
        Str(FireEnd:3:2,s);
        GSetColor(Black);
        SetTextStyle(MenuFont,HorizDir,MenuSize);
        SetTextJustify(CenterText,CenterText);
        If Direct=Horiz then Begin
           SetTextJustify(LeftText,BottomText);
           OutTextXY(Xpos - Size div 2 - 4, Ypos - 8, s);
          end{then}
        else Begin
           SetTextJustify(CenterText,BottomText);
           OutTextXY(Xpos, Ypos - Size div 2 - 6, s);
        end{if};
     end{if};
  end;

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

  Procedure PetriTransition.TErase;
  Begin
     EraseShape;
     EraseStatus;
  end;

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

  Function PetriTransition.Enabled;
  var n:Nodepoint;
      a:PetriAP;            { Arc }
      p:PetriPP;          { Place }
      en:Boolean;
  Begin
     If Empty(Inlist) then Enabled := false
     else Begin
      n := TakeFirst(Inlist);
      Repeat
        a := n^.el;
        p := a^.Place;
        en := p^.Marking >= a^.Weight; { This place OK or not }
        n := TakeNext(n);
      Until (n=nil) or not en;
      Enabled := en;
     end{if};
  end;

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

  Procedure PetriTransition.StartFire;
  var n:Nodepoint;
      a:PetriAP;            { Arc }
      p:PetriPP;          { Place }
  Begin
      n := TakeFirst(Inlist);
      While n<>nil do Begin
        a := n^.el;
        p := a^.Place;
        Dec(p^.Marking,a^.Weight);
        n := TakeNext(n);
      end{while};
  end;

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

  Procedure PetriTransition.EndFire;
  var n:Nodepoint;
      a:PetriAP;            { Arc }
      p:PetriPP;          { Place }
  Begin
      n := TakeFirst(Outlist);
      While n<>nil do Begin
        a := n^.el;
        p := a^.Place;
        Inc(p^.Marking,a^.Weight);
        n := TakeNext(n);
      end{while};
  end;

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

  Procedure PetriTransition.Fire;
  Begin
      StartFire;
      EndFire;
  end;

{********************* Methods for PetriArc ***********************}

  Procedure PetriArc.AInit(Col:Byte; Wid:Byte;
                           Sty:Byte; Wei:LongInt; Typ:Integer);
  Begin
     Init(Col,Wid,Sty);
     Weight := Wei;
     AType := Typ;
     BranchPro := 0;
     BranchCdf := 0;
  end;

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

  Procedure PetriArc.ShowWeight;
  var m,n:Nodepoint;
      p,q:PPnt;
      x,y:Word;
  Begin
     If not Empty(Corners) and (Abs(Weight)>1) then Begin
         m := TakeFirst(Corners);
         n := TakeNext(m);
         If n<>nil then Begin
             p := m^.el; q := n^.el;
             x := (p^.Xp+q^.Xp) div 2;
             If p^.Xp=q^.Xp then
               If Abs(Weight)<10 then Dec(x,6) else Dec(x,10);
             y := (p^.Yp+q^.Yp) div 2;
             If p^.Yp=q^.Yp then Dec(y,5);
             ShowNum(Abs(Weight),x,y,MarCol)
         end{if}
     end{if}
  end;

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

  Procedure PetriArc.EraseWeight;
  var m,n:Nodepoint;
      p,q:PPnt;
      x,y:Word;
  Begin
     If not Empty(Corners) and (Abs(Weight)>1) then Begin
         m := TakeFirst(Corners);
         n := TakeNext(m);
         If n<>nil then Begin
             p := m^.el; q := n^.el;
             x := (p^.Xp+q^.Xp) div 2;
             If p^.Xp=q^.Xp then
               If Abs(Weight)<10 then Dec(x,6) else Dec(x,10);
             y := (p^.Yp+q^.Yp) div 2;
             If p^.Yp=q^.Yp then Dec(y,5);
             ShowNum(Abs(Weight),x,y,Black)
         end{if}
     end{if}
  end;

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

  Procedure PetriArc.AShow;
  Begin
     ShowWithArrow;
     ShowWeight;
  end;

{+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

  Procedure PetriArc.AErase;
  Begin
     EraseWithArrow;
     EraseWeight;
  end;

End.
