Unit Randoms;                    { Random numbers for PetriSim }

Interface

Type
 PointP = ^FXPointType;   { Pointers to F(x) points }
 FXPointType = Record
                 Px    : Real;            { The Pdf }
                 Fx    : Real;            { The Cdf }
                 Value : Real;          { The value }
                 Next  : PointP;
               end;

 FXTableType = PointP;

 { Random variables have their own random stream ! }

 Rand = Object                        { Random variable object }
             FXList : FXTableType;{ Distribution function list }
             Seed : LongInt;           { Random generator seed }
             InitSeed : LongInt;      { Initial seed, 0 = none }
             DistType : Integer;           { Distribution type }
             P1,P2,P3 : Real;          { Real dist. parameters }
             PI1 : Integer;          { Integer dist. parameter }

             Constructor Init;                { Initialisation }
             Function FxAdd(V,P : Real) : Boolean;
        { Adds a point to F(x) with all tests, Cdf set to 1 !! }
             Function RValue : Real;        { The random value }
          End{Rand};

 Function Uniform(First, Last : Real) : Real;
 Function Normal(Mean, Std : Real) : Real;
 Function NormalnSig(Mean, Std : Real; n : Integer) : Real;
 Function Exponential(Mean : Real) : Real;
 Function Erlang(Mean : Real; Shape : Integer) : Real;
 Function Triangular(MinV,MLV,MaxV : Real) : Real;

Implementation

{=========================== Rand =============================}

Constructor Rand.Init;          { Initialisation }
Begin
   FXList := nil;
   InitSeed := 0;                { 0 = Randomize }
   DistType := 0;          { Constant duration 0 }
   P1 := 0;
   P2 := 0;
   P3 := 0;
   PI1 := 0;
End;

Function Rand.FxAdd(V,P : Real) : Boolean;
        { Adds a point to F(x) with all tests, Cdf set to 1 !! }
Var fxp,n : PointP;
    x     : Real;
Begin
  FxAdd := false;
  If (P>=0) and (P<=1) then begin
    If FXList=nil then begin
       new(fxp);
       fxp^.Fx := 1;
       fxp^.Next := nil;
       fxp^.Value := V;
       fxp^.Px := P;
       FXList := fxp;
       FxAdd := true;
     end
    else begin
       n := FXList;
       x := 0;
       while n^.Next<>nil do begin
         x := x + n^.Px;
         {n^.Fx := x;}
         n := n^.Next;
       end;
       x := x + n^.Px;    { Last point }
       If V>=n^.Value then begin
         if x>1 then n^.Fx := 1
                else n^.Fx := x;
         new(fxp);
         fxp^.Fx := 1;
         fxp^.Next := nil;
         fxp^.Value := V;
         fxp^.Px := P;
         n^.Next := fxp;
         FxAdd := true;
       end;
    end;
  end;
End;

Function Rand.RValue : Real;
Var Temp : LongInt;
    X,Pr1,Pr2,V1,V2 : Real;
    Pre,Suc : PointP;
Begin
  Temp := RandSeed;         { Saving current seed }
  RandSeed := Seed;            { Using local seed }
  Case DistType of

  0 : RValue := P1;                    { Constant }

  1,2 :                         { Empirical table }
      If FXList=nil then
        RValue := 0
      else begin
        X := Random;
        Pre := nil;      { Searching in F(x) }
        Suc := FXList;
        while Suc^.Fx<X do begin
          Pre := Suc;
          Suc := Suc^.Next;
        end;
        if DistType=1 then        { Discrete }
          RValue := Suc^.Value
        else                    { Continuous }
          { Linear Interpolation: }
          if Pre<>nil then begin
            V1 := Pre^.Value;
            Pr1 := Pre^.Fx;
           end
          else begin
            V1 := 0; Pr1 := 0
          end;
          V2 := Suc^.Value;
          Pr2 := Suc^.Fx;
          RValue := V1 + (V2-V1)*(X-Pr1)/(Pr2-Pr1);
      end;

  3 : RValue := Exponential(P1);    { Exponential }

  4 : RValue := Erlang(P1,PI1);          { Erlang }

  5 : RValue := Uniform(P1,P2);         { Uniform }

  6 : RValue := Triangular(P1,P2,P3);{ Triangular }

  7 : RValue := Normal(P1,P2);           { Normal }

  8 : RValue := NormalnSig(P1,P2,PI1);  { Normal in n Sigma }
  end;
  Seed := RandSeed;           { Saving local seed }
  RandSeed := Temp;      { Restoring current seed }
End;

{========================== Functions =============================}

 Function Uniform(First, Last : Real) : Real;
 Begin
    Uniform := First + Random*(Last - First);
 End;

 Function Normal(Mean, Std: Real) : Real;
 Var x : Real;
     i : Integer;
 Begin
    x := 0;
    For i:=1 to 12 do x := x + Random;
    Normal := (x - 6)*Std + Mean;
 End;

 Function NormalnSig(Mean, Std: Real; n : Integer) : Real;
 { "Normal" truncated to +/- n sigma }
 Var x : Real;
     i : Integer;
 Begin
   Repeat
    x := 0;
    For i:=1 to 12 do x := x + Random;
   Until Abs(x-6)<=n;
   NormalnSig := (x - 6)*Std + Mean;
 End;

 Function Exponential(Mean: Real) : Real;
 Begin
    If Mean<=0 then Exponential := 0
    Else Exponential := -Mean * Ln(Random);
 End;

 Function Erlang(Mean: Real; Shape: Integer) : Real;
 Var ExpMean,x : Real;
     i : Integer;
 Begin
    ExpMean := Mean/Shape;
    x := 0;
    For i:=1 to Shape do x := x + Exponential(ExpMean);
    Erlang := x;
 End;

 Function Triangular(MinV,MLV,MaxV: Real) : Real;
 Var u,x : Real;
     i : Integer;
     found : Boolean;
 Begin
    Repeat                   { Rejection method }
      x := Uniform(MinV, MaxV);
      u := Random;
      if x<MLV then                { Left slope }
        found := (u <= (x - MinV)/(MLV - MinV))
      else if x>MLV then          { Right slope }
        found := (u <= (MaxV - x)/(MaxV - MLV))
      else                     { Exactly middle }
        found := true;
    until found;
    Triangular := x;
 End;

End.


