Unit Statist;      { Statistics support for PetriSim }

Interface

Uses Simpet5;          { Needs the model time PNTime }

Type
 PRAccum = ^RAccum;
 PRTally = ^RTally;

 RAccum = Object { Statistically observed real number, time accumulation }
           Private
             X : Real;                        { The real value }
             Max : Real;                { So far maximum value }
             Min : Real;                { So far minimum value }
             Sum : Real;                        { The integral }
             SumSq : Real;           { The integral of squares }
             InitTime : Real;            { Initialisation time }
             LastTime : Real;               { Last update time }
             Empty : Boolean;                  { Initial state }
           Public
             Constructor Init(XInit : Real);
              { Statistics Initialisation, storing first value }
             Procedure RUpdate(XNew : Real);
                      { Storing new value, updating statistics }
             Function RValue : Real;
                                   { Returns the current value }
             Function RAverage : Real;
                            { Returns the current average value.
                            Call RAverage AFTER LAST UPDATE !  }
             Function RMax : Real;
                            { Returns the so far maximum value }
             Function RMin : Real;
                            { Returns the so far minimum value }
             Function RVariance: Real;
                                 { Returns the current variance.
                           Call RVariance AFTER LAST UPDATE !  }
             Function RStdDev: Real;
                       { Returns the current standard deviation.
                             Call RStdDev AFTER LAST UPDATE !  }
          End{RAccum};

 RTally = Object
           Private
             X : Real;                        { The real value }
             Max : Real;                { So far maximum value }
             Min : Real;                { So far minimum value }
             Sum : Real;                        { The integral }
             SumSq : Real;           { The integral of squares }
             UpdateCount : LongInt;        { Number of updates }
             Empty : Boolean;                  { Initial state }
           Public
             Constructor Init;
                                   { Statistics Initialisation }
             Procedure RUpdate(XNew : Real);
                      { Storing new value, updating statistics }
             Function RValue : Real;
                                   { Returns the current value }
             Function RCount : Real;
                               { Returns the number of updates }
             Function RAverage : Real;
                            { Returns the current average value.
                            Call RAverage AFTER LAST UPDATE !  }
             Function RMax : Real;
                            { Returns the so far maximum value }
             Function RMin : Real;
                            { Returns the so far minimum value }
             Function RVariance: Real;
                                 { Returns the current variance.
                           Call RVariance AFTER LAST UPDATE !  }
             Function RStdDev: Real;
                       { Returns the current standard deviation.
                             Call RStdDev AFTER LAST UPDATE !  }
          End{RTally};

Implementation

{========================= RAccum =============================}

Constructor RAccum.Init(XInit : Real);
Begin
   X   := XInit;
   Max := XInit;
   Min := XInit;
   Sum := 0;
   SumSq := 0;
   InitTime := PNTime;
   LastTime := InitTime;
End;

Procedure RAccum.RUpdate(XNew : Real);
Var d : Real;
Begin
   d := (PNTime-LastTime)*X;
   Sum := Sum + d;
   SumSq := SumSq + d*X;
   LastTime := PNTime;
   X := XNew;
   If XNew>Max then Max := XNew;
   If XNew<Min then Min := XNew;
End;

Function RAccum.RValue : Real;
                     { Returns the current value }
Begin
   RValue := X
End;

Function RAccum.RAverage : Real;
              { Returns the current average value.
              Call RAverage AFTER LAST UPDATE !  }
Begin
  If PNTime>InitTime then
    RAverage := Sum/(PNTime - InitTime)
  Else
    RAverage := 0
End;

Function RAccum.RMax : Real;
              { Returns the so far maximum value }
Begin
   RMax := Max
End;

Function RAccum.RMin : Real;
              { Returns the so far minimum value }
Begin
   RMin := Min
End;

Function RAccum.RVariance: Real;
                  { Returns the current variance }
Begin
  If PNTime>InitTime then
    RVariance := SumSq/(PNTime-InitTime) - Sqr(RAverage)
  Else
    RVariance := 0
End;

Function RAccum.RStdDev: Real;
        { Returns the current standard deviation }
Begin
   RStdDev := Sqrt(Abs(RVariance))
End;

{========================= RTally =============================}

Constructor RTally.Init;
Begin
   Empty := true;
   X := 0;                { To return if empty }
   Max := 0;
   Min := 0;
End;

Procedure RTally.RUpdate(XNew : Real);
Begin
 If Empty then begin
   Sum := XNew;
   SumSq := XNew*XNew;
   X := XNew;
   Max := XNew;
   Min := XNew;
   UpDateCount := 1;
   Empty := false;
  end{then}
 Else begin
   Inc(UpDateCount);
   Sum := Sum + XNew;
   SumSq := SumSq + XNew*XNew;
   X := XNew;
   If XNew>Max then Max := XNew;
   If XNew<Min then Min := XNew;
 End;
End;

Function RTally.RValue : Real;
       { Returns the current value (0 for empty) }
Begin
   RValue := X
End;

Function RTally.RCount : Real;
                 { Returns the number of updates }
Begin
  If Empty then
    RCount := 0
  Else
    RCount := UpDateCount
End;

Function RTally.RAverage : Real;
             { Returns the current average value }
Begin
  If not Empty then
    RAverage := Sum/UpDateCount
  Else
    RAverage := 0
End;

Function RTally.RMax : Real;
              { Returns the so far maximum value }
Begin
   RMax := Max
End;

Function RTally.RMin : Real;
              { Returns the so far minimum value }
Begin
   RMin := Min
End;

Function RTally.RVariance: Real;
                  { Returns the current variance }
Begin
  If not Empty then
    RVariance := SumSq/UpDateCount - Sqr(RAverage)
  Else
    RVariance := 0
End;

Function RTally.RStdDev: Real;
        { Returns the current standard deviation }
Begin
   RStdDev := Sqrt(Abs(RVariance))
End;

End.
